import os
from typing import Final
from xecai.memory.memory_interface import MemoryInterface
from xecai.models import Conversation

try:
    from psycopg_pool import AsyncConnectionPool, ConnectionPool
except ImportError as e:
    raise RuntimeError(
        """PostgreSQL memory provider requires the 'postgresql' extra.
        Install with: uv pip install xecai[postgresql]"""
    ) from e


class PostgreSQLMemory(MemoryInterface):
    def __init__(self):
        self.pg_url: Final[str] = os.environ.get(
            "POSTGRESQL_URL", "postgresql://postgres:postgres@127.0.0.1:5432/xecai"
        )
        self._pool: ConnectionPool | None = None
        self._async_pool: AsyncConnectionPool | None = None
        self._ensure_table_exists()

    def _get_pool(self) -> ConnectionPool:
        if self._pool is None:
            self._pool = ConnectionPool(self.pg_url)
        return self._pool

    async def _get_async_pool(self) -> AsyncConnectionPool:
        if self._async_pool is None:
            self._async_pool = AsyncConnectionPool(self.pg_url, open=False)
            await self._async_pool.open()
        return self._async_pool

    def _ensure_table_exists(self) -> None:
        query = """
        CREATE TABLE IF NOT EXISTS conversations (
            id UUID PRIMARY KEY,
            data JSONB NOT NULL
        );
        """
        with self._get_pool().connection() as conn:
            with conn.cursor() as cur:
                cur.execute(query)
            conn.commit()

    def sync_get_conversation(self, conversation_id: str) -> Conversation | None:
        query = "SELECT data FROM conversations WHERE id = %s;"
        with self._get_pool().connection() as conn:
            with conn.cursor() as cur:
                cur.execute(query, (conversation_id,))
                result = cur.fetchone()
                if result:
                    data = result[0]
                    if isinstance(data, str):
                        return Conversation.model_validate_json(data)
                    return Conversation.model_validate(data)
        return None

    async def async_get_conversation(self, conversation_id: str) -> Conversation | None:
        query = "SELECT data FROM conversations WHERE id = %s;"

        # 2. Await the pool fetching
        pool = await self._get_async_pool()

        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                await cur.execute(query, (conversation_id,))
                result = await cur.fetchone()
                if result:
                    data = result[0]
                    if isinstance(data, str):
                        return Conversation.model_validate_json(data)
                    return Conversation.model_validate(data)
        return None

    def sync_save_conversation(self, conversation: Conversation) -> None:
        query = """
        INSERT INTO conversations (id, data)
        VALUES (%s, %s)
        ON CONFLICT (id) DO UPDATE SET data = EXCLUDED.data;
        """
        with self._get_pool().connection() as conn:
            with conn.cursor() as cur:
                cur.execute(
                    query,
                    (
                        str(conversation.conversation_id),
                        conversation.model_dump_json(),
                    ),
                )
            conn.commit()

    async def async_save_conversation(self, conversation: Conversation) -> None:
        query = """
        INSERT INTO conversations (id, data)
        VALUES (%s, %s)
        ON CONFLICT (id) DO UPDATE SET data = EXCLUDED.data;
        """

        # 2. Await the pool fetching
        pool = await self._get_async_pool()

        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                await cur.execute(
                    query,
                    (
                        str(conversation.conversation_id),
                        conversation.model_dump_json(),
                    ),
                )
            await conn.commit()



import os
from typing import Final
from xecai.embeddings.embedding_interface import EmbeddingInterface

try:
    import openai
except ImportError as e:
    raise RuntimeError(
        """OpenAI provider requires the 'openai' extra.
        Install with: uv pip install xecai[openai]"""
    ) from e


class OpenAIEmbedding(EmbeddingInterface):
    def __init__(self):
        api_key = os.environ.get("OPENAI_API_KEY")
        self.client: Final[openai.OpenAI] = openai.OpenAI(api_key=api_key)
        self.aclient: Final[openai.AsyncOpenAI] = openai.AsyncOpenAI(api_key=api_key)

    def sync_get_embeddings(self, text: str, model: str) -> list[float]:
        response = self.client.embeddings.create(input=text, model=model)
        return response.data[0].embedding if response.data else []

    async def async_get_embeddings(self, text: str, model: str) -> list[float]:
        response = await self.aclient.embeddings.create(input=text, model=model)
        return response.data[0].embedding if response.data else []



from abc import ABC, abstractmethod


class EmbeddingInterface(ABC):
    @abstractmethod
    def sync_get_embeddings(self, text: str, model: str) -> list[float]:
        pass

    @abstractmethod
    async def async_get_embeddings(self, text: str, model: str) -> list[float]:
        pass



from xecai.models import Message, MessageType


def trim_conversation(messages: list[Message], max_chars: int) -> list[Message]:
    result_messages: list[Message] = []
    result_chars = 0
    for msg in reversed(messages):
        msg_len = len(msg.content)
        if result_chars + msg_len > max_chars:
            break

        result_chars += msg_len
        result_messages.insert(0, msg)

    while result_messages and result_messages[0].message_type != MessageType.USER:
        result_messages.pop(0)

    return result_messages



import os
from collections.abc import AsyncIterator, Iterator
from typing import Final

from xecai.chat.chat_interface import ChatInterface
from xecai.models import (
    ChatResponse,
    Message,
    MessageType,
    ReasoningOptions,
    Stats,
    StopReason,
    ModelNotFoundError,
)

from xecai.error_handlers.google_error_handler import (
    async_error_decorator,
    async_generator_error_decorator,
    sync_error_decorator,
    sync_generator_error_decorator,
)

try:
    from google import genai
    from google.genai import types
except ImportError as e:
    raise RuntimeError(
        """Google provider requires the 'google' extra.
        Install with: uv pip install xecai[google]"""
    ) from e


class GoogleChat(ChatInterface):
    def __init__(self):
        self.client: Final[genai.Client] = genai.Client(
            api_key=os.environ.get("GOOGLE_API_KEY")
        )

    @sync_error_decorator
    def check_model(self, model_name: str) -> None:
        available_models = list(self.client.models.list())
        # Google model names usually have a "models/" prefix
        if not any(
            m.name == model_name or m.name == f"models/{model_name}"
            for m in available_models
        ):
            raise ModelNotFoundError(f"Model '{model_name}' not found.")

    @staticmethod
    def messages_to_custom_messages(
        system_prompt: str, messages: list[Message]
    ) -> list:
        inputs = []
        for msg in messages:
            if msg.message_type == MessageType.BOT:
                role = "model"
            else:
                role = "user"

            inputs.append({"role": role, "parts": [{"text": msg.content}]})
        return inputs

    def _prepare_config(
        self,
        system_prompt: str,
        reasoning: ReasoningOptions | None,
        temperature: float | None,
        retries: int = 3,
    ) -> types.GenerateContentConfig:
        kwargs = {"system_instruction": system_prompt}

        if temperature is not None:
            kwargs["temperature"] = temperature

        if reasoning and reasoning != ReasoningOptions.NONE:
            budget_map = {
                ReasoningOptions.LOW: 1024,
                ReasoningOptions.MEDIUM: 4096,
                ReasoningOptions.HIGH: 16384,
            }
            kwargs["thinking_config"] = {"thinking_budget": budget_map[reasoning]}

        return types.GenerateContentConfig(**kwargs)

    def _map_stop_reason(self, finish_reason: str | int | None) -> StopReason:
        if finish_reason is None:
            return StopReason.END

        # Convert to string to handle enum names or raw values
        reason_str = str(finish_reason).upper()

        if "STOP" in reason_str:
            return StopReason.END
        elif "MAX_TOKENS" in reason_str:
            return StopReason.MAX_TOKENS
        elif (
            "SAFETY" in reason_str
            or "BLOCKLIST" in reason_str
            or "PROHIBITED_CONTENT" in reason_str
            or "SPII" in reason_str
        ):
            return StopReason.CONTENT_FILTER
        elif "UNEXPECTED_TOOL_CALL" in reason_str:
            return StopReason.TOOL_USE

        return StopReason.OTHER

    @sync_error_decorator
    def invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        response = self.client.models.generate_content(
            model=model_name,
            contents=self.messages_to_custom_messages(system_prompt, messages),
            config=self._prepare_config(
                system_prompt, reasoning, temperature
            ),
        )

        usage = response.usage_metadata
        stats = (
            Stats(
                input_tokens=getattr(usage, "prompt_token_count", 0) or 0,
                output_tokens=getattr(usage, "candidates_token_count", 0) or 0,
                total_tokens=getattr(usage, "total_token_count", 0) or 0,
                cached_input_tokens=getattr(usage, "cached_content_token_count", 0)
                or 0,
                reasoning_tokens=getattr(usage, "thoughts_token_count", 0) or 0,
            )
            if usage
            else Stats()
        )

        candidates = getattr(response, "candidates", [])
        finish_reason = None
        if candidates and len(candidates) > 0:
            finish_reason = getattr(candidates[0], "finish_reason", None)

        return ChatResponse(
            text=response.text or "",
            stats=stats,
            stop_reason=self._map_stop_reason(finish_reason),
        )

    @async_error_decorator
    async def async_invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        response = await self.client.aio.models.generate_content(
            model=model_name,
            contents=self.messages_to_custom_messages(system_prompt, messages),
            config=self._prepare_config(
                system_prompt, reasoning, temperature
            ),
        )

        usage = response.usage_metadata
        stats = (
            Stats(
                input_tokens=getattr(usage, "prompt_token_count", 0) or 0,
                output_tokens=getattr(usage, "candidates_token_count", 0) or 0,
                total_tokens=getattr(usage, "total_token_count", 0) or 0,
                cached_input_tokens=getattr(usage, "cached_content_token_count", 0)
                or 0,
                reasoning_tokens=getattr(usage, "thoughts_token_count", 0) or 0,
            )
            if usage
            else Stats()
        )

        candidates = getattr(response, "candidates", [])
        finish_reason = None
        if candidates and len(candidates) > 0:
            finish_reason = getattr(candidates[0], "finish_reason", None)

        return ChatResponse(
            text=response.text or "",
            stats=stats,
            stop_reason=self._map_stop_reason(finish_reason),
        )

    @sync_generator_error_decorator
    def stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> Iterator[ChatResponse]:
        response = self.client.models.generate_content_stream(
            model=model_name,
            contents=self.messages_to_custom_messages(system_prompt, messages),
            config=self._prepare_config(
                system_prompt, reasoning, temperature
            ),
        )

        final_stats = None
        finish_reason = None
        for chunk in response:
            if chunk.usage_metadata:
                usage = chunk.usage_metadata
                final_stats = Stats(
                    input_tokens=getattr(usage, "prompt_token_count", 0) or 0,
                    output_tokens=getattr(usage, "candidates_token_count", 0) or 0,
                    total_tokens=getattr(usage, "total_token_count", 0) or 0,
                    cached_input_tokens=getattr(usage, "cached_content_token_count", 0)
                    or 0,
                    reasoning_tokens=getattr(usage, "thoughts_token_count", 0) or 0,
                )

            candidates = getattr(chunk, "candidates", [])
            if candidates and len(candidates) > 0:
                chunk_reason = getattr(candidates[0], "finish_reason", None)
                if (
                    chunk_reason is not None
                    and str(chunk_reason) != "FINISH_REASON_UNSPECIFIED"
                ):
                    finish_reason = chunk_reason

            if chunk.text:
                yield ChatResponse(text=chunk.text)

        if final_stats or finish_reason is not None:
            yield ChatResponse(
                text="",
                stats=final_stats,
                stop_reason=self._map_stop_reason(finish_reason),
            )

    @async_generator_error_decorator
    async def async_stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> AsyncIterator[ChatResponse]:
        response = await self.client.aio.models.generate_content_stream(
            model=model_name,
            contents=self.messages_to_custom_messages(system_prompt, messages),
            config=self._prepare_config(
                system_prompt, reasoning, temperature
            ),
        )

        final_stats = None
        finish_reason = None
        async for chunk in response:
            if chunk.usage_metadata:
                usage = chunk.usage_metadata
                final_stats = Stats(
                    input_tokens=getattr(usage, "prompt_token_count", 0) or 0,
                    output_tokens=getattr(usage, "candidates_token_count", 0) or 0,
                    total_tokens=getattr(usage, "total_token_count", 0) or 0,
                    cached_input_tokens=getattr(usage, "cached_content_token_count", 0)
                    or 0,
                    reasoning_tokens=getattr(usage, "thoughts_token_count", 0) or 0,
                )

            candidates = getattr(chunk, "candidates", [])
            if candidates and len(candidates) > 0:
                chunk_reason = getattr(candidates[0], "finish_reason", None)
                if (
                    chunk_reason is not None
                    and str(chunk_reason) != "FINISH_REASON_UNSPECIFIED"
                ):
                    finish_reason = chunk_reason

            if chunk.text:
                yield ChatResponse(text=chunk.text)

        if final_stats or finish_reason is not None:
            yield ChatResponse(
                text="",
                stats=final_stats,
                stop_reason=self._map_stop_reason(finish_reason),
            )



from xecai.chat.chat_interface import ChatInterface
from xecai.models import Message, MessageType

DEFAULT_CONDENSE_PROMPT = """
    You are a summary assistant.
    You will receive a conversation between a human and a bot, and a new question.
    Your task is to create a new, independent question that takes all of the information
    from the context and does not miss key information needed in the question.
    """


def _prepare_condense_payload(
    previous_messages: list[Message], new_question: str
) -> list[Message] | None:
    """Prepares the message payload for the LLM. Returns None if no context exists."""
    if not previous_messages:
        return None

    conversation_text = "\n".join([msg.to_prompt_text() for msg in previous_messages])
    condense_content = f"""
        Conversation:

        {conversation_text}

        Question: {new_question}
        """
    return [Message(message_type=MessageType.USER, content=condense_content)]


def sync_condense_question(
    condenser_chat: ChatInterface,
    condenser_model_name: str,
    previous_messages: list[Message],
    new_question: str,
    CONDENSE_PROMPT: str = DEFAULT_CONDENSE_PROMPT,
) -> str:
    messages = _prepare_condense_payload(previous_messages, new_question)
    if not messages:
        return new_question

    response = condenser_chat.invoke(condenser_model_name, CONDENSE_PROMPT, messages)
    return response.text


async def async_condense_question(
    condenser_chat: ChatInterface,
    condenser_model_name: str,
    previous_messages: list[Message],
    new_question: str,
    CONDENSE_PROMPT: str = DEFAULT_CONDENSE_PROMPT,
) -> str:
    messages = _prepare_condense_payload(previous_messages, new_question)
    if not messages:
        return new_question

    response = await condenser_chat.async_invoke(
        condenser_model_name, CONDENSE_PROMPT, messages
    )
    return response.text



import os
from typing import Final
from xecai.embeddings.embedding_interface import EmbeddingInterface

try:
    from google import genai
except ImportError as e:
    raise RuntimeError(
        """Google provider requires the 'google' extra.
        Install with: uv pip install xecai[google]"""
    ) from e


class GoogleEmbedding(EmbeddingInterface):
    def __init__(self):
        self.client: Final[genai.Client] = genai.Client(
            api_key=os.environ.get("GOOGLE_API_KEY")
        )

    def sync_get_embeddings(self, text: str, model: str) -> list[float]:
        response = self.client.models.embed_content(
            model=model,
            contents=text,
        )
        return response.embeddings[0].values if response.embeddings else []

    async def async_get_embeddings(self, text: str, model: str) -> list[float]:
        response = await self.client.aio.models.embed_content(
            model=model,
            contents=text,
        )
        return response.embeddings[0].values if response.embeddings else []



from xecai.models import Chunk


def make_rag_prompt(query: str, chunks: list[Chunk]) -> str:
    context_str_parts: list[str] = []
    for chunk in chunks:
        part = f"Document: {chunk.document}\nContent: {chunk.content}"
        if chunk.score is not None:
            part += f"\nScore: {chunk.score}"
        context_str_parts.append(part)
    context_str = "\n\n".join(context_str_parts)
    return f"Context information:\n{context_str}\n\nGiven the context information, answer the following query: {query}"



import functools
from collections.abc import AsyncIterator, Awaitable, Iterator
from typing import Callable, NoReturn, ParamSpec, TypeVar
from xecai.models import CredentialsError, BadRequestError, RateLimitError

try:
    import botocore.exceptions
except ImportError as e:
    raise RuntimeError(
        """AWS Bedrock provider requires the 'aws' extra.
        Install with: uv pip install xecai[aws]"""
    ) from e

P = ParamSpec("P")
T = TypeVar("T")


def handle_client_error(e: botocore.exceptions.ClientError) -> NoReturn:
    error_code = e.response.get("Error", {}).get("Code", "")
    message = e.response.get("Error", {}).get("Message", str(e))

    if error_code == "ThrottlingException" or error_code == "InternalServerException":
        raise RateLimitError(message) from e
    elif error_code == "ValidationException":
        raise BadRequestError(message) from e
    elif error_code in ("UnrecognizedClientException", "AccessDeniedException"):
        raise CredentialsError(message) from e

    raise e


def handle_botocore_error(e: botocore.exceptions.BotoCoreError) -> NoReturn:
    if isinstance(e, botocore.exceptions.NoCredentialsError):
        raise CredentialsError(str(e)) from e

    raise e


def sync_error_decorator(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return func(*args, **kwargs)
        except botocore.exceptions.ClientError as e:
            handle_client_error(e)
        except botocore.exceptions.BotoCoreError as e:
            handle_botocore_error(e)

    return wrapper


def async_error_decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return await func(*args, **kwargs)
        except botocore.exceptions.ClientError as e:
            handle_client_error(e)
        except botocore.exceptions.BotoCoreError as e:
            handle_botocore_error(e)

    return wrapper


def sync_generator_error_decorator(
    func: Callable[P, Iterator[T]],
) -> Callable[P, Iterator[T]]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Iterator[T]:
        try:
            yield from func(*args, **kwargs)
        except botocore.exceptions.ClientError as e:
            handle_client_error(e)
        except botocore.exceptions.BotoCoreError as e:
            handle_botocore_error(e)

    return wrapper


def async_generator_error_decorator(
    func: Callable[P, AsyncIterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> AsyncIterator[T]:
        try:
            async for item in func(*args, **kwargs):
                yield item
        except botocore.exceptions.ClientError as e:
            handle_client_error(e)
        except botocore.exceptions.BotoCoreError as e:
            handle_botocore_error(e)

    return wrapper



import pytest
from xecai.utils.trim_conversation import trim_conversation
from xecai.utils.rag_prompt import make_rag_prompt
from xecai.utils.condenser import sync_condense_question
from xecai.utils.summarizer import _prepare_summary_payload, sync_summarize_conversation
from xecai.models import Message, MessageType, Chunk, Conversation

def test_trim_conversation_basic():
    messages = [Message(content="hello", message_type=MessageType.USER), Message(content="world", message_type=MessageType.BOT)]
    trimmed = trim_conversation(messages, max_chars=20)
    assert isinstance(trimmed, list)
    assert len(trimmed) == 2
    assert trimmed[0].content == "hello"
    assert trimmed[1].content == "world"

def test_trim_conversation_with_max_chars():
    messages = [Message(content="hello", message_type=MessageType.USER), Message(content="world", message_type=MessageType.BOT)]
    trimmed = trim_conversation(messages, max_chars=5)
    assert isinstance(trimmed, list)
    assert len(trimmed) == 0

def test_make_rag_prompt():
    chunk = Chunk(document="doc1", content="some content", score=0.9)
    prompt = make_rag_prompt("What is this?", [chunk])
    assert "doc1" in prompt
    assert "some content" in prompt
    assert "Score: 0.9" in prompt
    assert "What is this?" in prompt

def test_sync_condense_question_returns_new_question(monkeypatch):
    class DummyChat:
        def invoke(self, model, prompt, messages):
            class Response:
                text = "condensed question"
            return Response()
    previous = [Message(content="context", message_type=MessageType.USER)]
    new_question = "What is up?"
    result = sync_condense_question(DummyChat(), "model", previous, new_question)
    assert result == "condensed question"

def test_prepare_summary_payload():
    conv = Conversation(messages=[Message(content="hello", message_type=MessageType.USER)])
    payload = _prepare_summary_payload(conv)
    assert isinstance(payload, list)
    assert payload[0].content == "USER: hello"

def test_sync_summarize_conversation(monkeypatch):
    class DummyMemory:
        def sync_get_conversation(self, conversation_id):
            return Conversation(messages=[Message(content="hi", message_type=MessageType.USER)])
        def sync_save_conversation(self, conversation):
            self.saved = conversation
    class DummyChat:
        def invoke(self, model_name, system_prompt, messages):
            class Response:
                text = "summary"
            return Response()
    memory = DummyMemory()
    chat = DummyChat()
    sync_summarize_conversation(memory, chat, "id", "model")
    assert hasattr(memory, "saved")
    assert memory.saved.messages[0].content == "summary"



import os
import uuid
from typing import Final
from xecai.models import Chunk, SearchType
from xecai.vector_db.vector_db_interface import VectorDBInterface

try:
    import pinecone
except ImportError as e:
    raise RuntimeError(
        """Pinecone vector DB provider requires the 'pinecone' extra.
        Install with: uv pip install xecai[pinecone]"""
    ) from e


class PineconeVectorDB(VectorDBInterface):
    """
    Pinecone vector database implementation.

    Pinecone is a managed vector database service. It requires an API key and
    an index name. This implementation uses Pinecone's Integrated Inference
    so you do not need to provide an embedding interface.
    """

    def __init__(
        self,
        index_name: str = "index",
        namespace: str = "",
    ):
        self.api_key: Final[str] = os.environ.get("PINECONE_API_KEY", "")
        self.index_name: Final[str] = os.environ.get("PINECONE_INDEX_NAME", index_name)
        # Using "__default__" for empty namespace per new SDK requirements
        self.namespace: Final[str] = namespace if namespace else "__default__"

        if not self.api_key:
            raise ValueError("PINECONE_API_KEY environment variable is not set.")

        self.pc = pinecone.Pinecone(api_key=self.api_key)

        # Check if index exists
        if self.index_name not in self.pc.list_indexes().names():
            raise ValueError(
                f"Pinecone index '{self.index_name}' does not exist. Please create it in the Pinecone console or via the API."
            )

        self.index = self.pc.Index(self.index_name)

    def sync_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> list[Chunk]:
        # Using integrated inference search
        response = self.index.search(
            namespace=self.namespace, query={"inputs": {"text": query}, "top_k": k}
        )

        results = []
        hits = response.get("result", {}).get("hits", [])
        for hit in hits:
            fields = hit.get("fields", {})
            results.append(
                Chunk(
                    document=fields.get("document", ""),
                    origin=fields.get("origin", ""),
                    fragment=int(fields.get("fragment", 0)),
                    content=fields.get("text", ""),
                    score=hit.get("_score"),
                )
            )

        return results

    async def async_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> list[Chunk]:
        # Pinecone SDK is primarily synchronous
        return self.sync_retrieve(query, k, search_type)

    def sync_insert(self, chunks: list[Chunk]) -> None:
        records_to_upsert = []
        for chunk in chunks:
            record_id = str(uuid.uuid4())

            record = {
                "_id": record_id,
                "text": chunk.content,
                "document": chunk.document,
                "origin": chunk.origin or "",
                "fragment": chunk.fragment,
            }

            records_to_upsert.append(record)

        # Upsert in batches of 96 as recommended for integrated inference models
        batch_size = 96
        for i in range(0, len(records_to_upsert), batch_size):
            batch = records_to_upsert[i : i + batch_size]
            self.index.upsert_records(namespace=self.namespace, records=batch)

    async def async_insert(self, chunks: list[Chunk]) -> None:
        self.sync_insert(chunks)

    def sync_get_num_documents(self) -> int:
        stats = self.index.describe_index_stats()
        if self.namespace:
            return (
                stats.get("namespaces", {})
                .get(self.namespace, {})
                .get("vector_count", 0)
            )
        return stats.get("total_vector_count", 0)

    async def async_get_num_documents(self) -> int:
        return self.sync_get_num_documents()



import os
from collections.abc import AsyncIterator, Iterator
from typing import Final
from xecai.chat.chat_interface import ChatInterface
from xecai.models import (
    ChatResponse,
    Message,
    MessageType,
    ReasoningOptions,
    Stats,
    StopReason,
    ModelNotFoundError,
)
from xecai.error_handlers.anthropic_error_handler import (
    async_error_decorator,
    async_generator_error_decorator,
    sync_error_decorator,
    sync_generator_error_decorator,
)

try:
    from anthropic import Anthropic, AsyncAnthropic
except ImportError as e:
    raise RuntimeError(
        """Anthropic provider requires the 'anthropic' extra.
        Install with: uv pip install xecai[anthropic]"""
    ) from e


class AnthropicChat(ChatInterface):
    def __init__(self):
        api_key = os.environ.get("ANTHROPIC_API_KEY")
        self.client: Final[Anthropic] = Anthropic(api_key=api_key)
        self.aclient: Final[AsyncAnthropic] = AsyncAnthropic(api_key=api_key)

    @sync_error_decorator
    def check_model(self, model_name: str) -> None:
        available_models = self.client.models.list()
        if not any(m.id == model_name for m in available_models.data):
            raise ModelNotFoundError(f"Model '{model_name}' not found.")

    @staticmethod
    def messages_to_custom_messages(
        system_prompt: str, messages: list[Message]
    ) -> list:
        contents = []
        for msg in messages:
            if msg.message_type == MessageType.BOT:
                role = "assistant"
            else:
                role = "user"

            contents.append({"role": role, "content": msg.content})
        return contents

    def _prepare_kwargs(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None,
        temperature: float | None,
        stream: bool = False,
        retries: int = 3,
    ) -> dict:

        kwargs = {
            "model": model_name,
            "max_tokens": 4096,
            "messages": self.messages_to_custom_messages(system_prompt, messages),
            "system": system_prompt,
        }

        if stream:
            kwargs["stream"] = True

        if temperature is not None:
            kwargs["temperature"] = temperature

        if reasoning and reasoning != ReasoningOptions.NONE:
            budget_map = {
                ReasoningOptions.LOW: 1024,
                ReasoningOptions.MEDIUM: 4096,
                ReasoningOptions.HIGH: 16384,
            }
            budget = budget_map[reasoning]
            kwargs["thinking"] = {"type": "enabled", "budget_tokens": budget}
            kwargs["max_tokens"] = budget + 4096
            kwargs["temperature"] = 1.0

        return kwargs

    def _map_stop_reason(self, stop_reason: str | None) -> StopReason:
        if stop_reason == "end_turn":
            return StopReason.END
        elif stop_reason == "max_tokens":
            return StopReason.MAX_TOKENS
        elif stop_reason == "stop_sequence":
            return StopReason.STOP_SEQUENCE
        elif stop_reason == "tool_use":
            return StopReason.TOOL_USE
        elif stop_reason == "content_filtered":
            return StopReason.CONTENT_FILTER
        elif stop_reason is None:
            return StopReason.END
        return StopReason.OTHER

    @sync_error_decorator
    def invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        response = self.client.messages.create(**kwargs)
        usage = response.usage
        stats = (
            Stats(
                input_tokens=getattr(usage, "input_tokens", 0) or 0,
                output_tokens=getattr(usage, "output_tokens", 0) or 0,
                total_tokens=(getattr(usage, "input_tokens", 0) or 0)
                + (getattr(usage, "output_tokens", 0) or 0),
                cached_input_tokens=getattr(usage, "cache_read_input_tokens", 0) or 0,
                reasoning_tokens=0,
            )
            if usage
            else Stats()
        )
        text = ""
        if response.content:
            for block in response.content:
                if block.type == "text":
                    text += block.text

        return ChatResponse(
            text=text,
            stats=stats,
            stop_reason=self._map_stop_reason(getattr(response, "stop_reason", None)),
        )

    @async_error_decorator
    async def async_invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        response = await self.aclient.messages.create(**kwargs)
        usage = response.usage
        stats = (
            Stats(
                input_tokens=getattr(usage, "input_tokens", 0) or 0,
                output_tokens=getattr(usage, "output_tokens", 0) or 0,
                total_tokens=(getattr(usage, "input_tokens", 0) or 0)
                + (getattr(usage, "output_tokens", 0) or 0),
                cached_input_tokens=getattr(usage, "cache_read_input_tokens", 0) or 0,
                reasoning_tokens=0,
            )
            if usage
            else Stats()
        )
        text = ""
        if response.content:
            for block in response.content:
                if block.type == "text":
                    text += block.text

        return ChatResponse(
            text=text,
            stats=stats,
            stop_reason=self._map_stop_reason(getattr(response, "stop_reason", None)),
        )

    @sync_generator_error_decorator
    def stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> Iterator[ChatResponse]:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
            stream=True,
        )
        response = self.client.messages.create(**kwargs)

        input_tokens = 0
        output_tokens = 0
        cached_tokens = 0
        stop_reason = None

        for event in response:
            if event.type == "message_start":
                usage = getattr(event.message, "usage", None)
                if usage:
                    input_tokens += getattr(usage, "input_tokens", 0) or 0
                    cached_tokens += getattr(usage, "cache_read_input_tokens", 0) or 0
            elif event.type == "message_delta":
                usage = getattr(event, "usage", None)
                if usage:
                    output_tokens += getattr(usage, "output_tokens", 0) or 0
                if hasattr(event.delta, "stop_reason"):
                    stop_reason = event.delta.stop_reason
            elif (
                event.type == "content_block_delta" and event.delta.type == "text_delta"
            ):
                yield ChatResponse(text=event.delta.text)
            elif event.type == "message_stop":
                stats = Stats(
                    input_tokens=input_tokens,
                    output_tokens=output_tokens,
                    total_tokens=input_tokens + output_tokens,
                    cached_input_tokens=cached_tokens,
                    reasoning_tokens=0,
                )
                yield ChatResponse(
                    text="", stats=stats, stop_reason=self._map_stop_reason(stop_reason)
                )

    @async_generator_error_decorator
    async def async_stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> AsyncIterator[ChatResponse]:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
            stream=True,
        )
        response = await self.aclient.messages.create(**kwargs)

        input_tokens = 0
        output_tokens = 0
        cached_tokens = 0
        stop_reason = None

        async for event in response:
            if event.type == "message_start":
                usage = getattr(event.message, "usage", None)
                if usage:
                    input_tokens += getattr(usage, "input_tokens", 0) or 0
                    cached_tokens += getattr(usage, "cache_read_input_tokens", 0) or 0
            elif event.type == "message_delta":
                usage = getattr(event, "usage", None)
                if usage:
                    output_tokens += getattr(usage, "output_tokens", 0) or 0
                if hasattr(event.delta, "stop_reason"):
                    stop_reason = event.delta.stop_reason
            elif (
                event.type == "content_block_delta" and event.delta.type == "text_delta"
            ):
                yield ChatResponse(text=event.delta.text)
            elif event.type == "message_stop":
                stats = Stats(
                    input_tokens=input_tokens,
                    output_tokens=output_tokens,
                    total_tokens=input_tokens + output_tokens,
                    cached_input_tokens=cached_tokens,
                    reasoning_tokens=0,
                )
                yield ChatResponse(
                    text="", stats=stats, stop_reason=self._map_stop_reason(stop_reason)
                )



from abc import ABC, abstractmethod
from xecai.models import Chunk


class RerankerInterface(ABC):
    @abstractmethod
    def sync_rerank(self, query: str, chunks: list[Chunk], k: int) -> list[Chunk]:
        pass

    @abstractmethod
    async def async_rerank(
        self, query: str, chunks: list[Chunk], k: int
    ) -> list[Chunk]:
        pass



import logging
import math
import re
from collections import Counter
from typing import List
from xecai.embeddings.embedding_interface import EmbeddingInterface
from xecai.models import Chunk, SearchType
from xecai.vector_db.vector_db_interface import VectorDBInterface


def tokenize(text: str) -> List[str]:
    """Simple alphanumeric tokenizer."""
    return re.findall(r"\w+", text.lower())


def compute_bm25_scores(
    query: str, documents: List[str], k1: float = 1.5, b: float = 0.75
) -> List[float]:
    """
    Computes BM25 scores for a query against a list of documents.
    """
    query_terms = tokenize(query)
    if not query_terms or not documents:
        return [0.0] * len(documents)

    doc_tokens = [tokenize(doc) for doc in documents]
    doc_lengths = [len(tokens) for tokens in doc_tokens]
    avgdl = sum(doc_lengths) / len(documents) if documents else 1.0
    N = len(documents)

    # Calculate document frequencies for query terms
    df = Counter()
    for tokens in doc_tokens:
        unique_tokens = set(tokens)
        for term in query_terms:
            if term in unique_tokens:
                df[term] += 1

    # Calculate IDF for query terms
    idf = {}
    for term in query_terms:
        # Standard BM25 IDF formula
        n_q = df[term]
        idf[term] = math.log(((N - n_q + 0.5) / (n_q + 0.5)) + 1.0)

    # Calculate scores
    scores = []
    for i, tokens in enumerate(doc_tokens):
        score = 0.0
        term_counts = Counter(tokens)
        doc_len = doc_lengths[i]

        for term in query_terms:
            if term not in term_counts:
                continue
            tf = term_counts[term]
            # BM25 term frequency normalization
            numerator = tf * (k1 + 1)
            denominator = tf + k1 * (1 - b + b * (doc_len / avgdl))
            score += idf[term] * (numerator / denominator)

        scores.append(score)

    return scores


def cosine_similarity(v1: List[float], v2: List[float]) -> float:
    dot_product = sum(a * b for a, b in zip(v1, v2))
    norm_v1 = math.sqrt(sum(a * a for a in v1))
    norm_v2 = math.sqrt(sum(b * b for b in v2))
    if norm_v1 == 0 or norm_v2 == 0:
        return 0.0
    return dot_product / (norm_v1 * norm_v2)


class DebugVectorDB(VectorDBInterface):
    """
    Debug implementation of VectorDBInterface that stores chunks in a list and queries them in-memory.
    Supports basic semantic search using Cosine Similarity and Hybrid Search combining it with BM25 keyword search.
    """

    def __init__(
        self,
        embedding_interface: EmbeddingInterface | None = None,
        embedding_model: str | None = None,
    ):
        if not embedding_interface or not embedding_model:
            raise ValueError(
                "embedding_interface and embedding_model must be provided."
            )

        logging.warning(
            "DebugVectorDB is for debugging purposes only and is not recommended for production usecases."
        )
        self.embedding_interface = embedding_interface
        self.embedding_model = embedding_model
        self._chunks: List[Chunk] = []
        self._embeddings: List[List[float]] = []

    def _search(
        self, query: str, query_embedding: List[float], k: int, search_type: SearchType
    ) -> List[Chunk]:
        if not self._chunks:
            return []

        # Calculate semantic scores
        semantic_scores = [
            cosine_similarity(query_embedding, chunk_emb)
            for chunk_emb in self._embeddings
        ]

        if search_type == SearchType.SEMANTIC:
            final_scores = semantic_scores
        else:
            # Hybrid search: Combine semantic score with BM25 keyword score
            documents = [chunk.content for chunk in self._chunks]
            bm25_scores = compute_bm25_scores(query, documents)

            # Normalize BM25 scores (min-max normalization) to combine with cosine similarity (0 to 1)
            max_bm25 = max(bm25_scores) if bm25_scores else 0
            min_bm25 = min(bm25_scores) if bm25_scores else 0

            normalized_bm25 = []
            for score in bm25_scores:
                if max_bm25 == min_bm25:
                    normalized_bm25.append(0.0 if max_bm25 == 0 else 1.0)
                else:
                    normalized_bm25.append((score - min_bm25) / (max_bm25 - min_bm25))

            # Simple average for hybrid
            final_scores = [
                (sem_score + kw_score) / 2.0
                for sem_score, kw_score in zip(semantic_scores, normalized_bm25)
            ]

        # Apply scores and sort
        scored_chunks = []
        for i, chunk in enumerate(self._chunks):
            result_chunk = chunk.model_copy()
            result_chunk.score = final_scores[i]
            scored_chunks.append(result_chunk)

        # Sort descending by score
        scored_chunks.sort(key=lambda x: x.score or 0.0, reverse=True)
        return scored_chunks[:k]

    def sync_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> List[Chunk]:
        query_embedding = self.embedding_interface.sync_get_embeddings(
            query, self.embedding_model
        )
        return self._search(query, query_embedding, k, search_type)

    async def async_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> List[Chunk]:
        query_embedding = await self.embedding_interface.async_get_embeddings(
            query, self.embedding_model
        )
        return self._search(query, query_embedding, k, search_type)

    def sync_insert(self, chunks: List[Chunk]) -> None:
        for chunk in chunks:
            embedding = self.embedding_interface.sync_get_embeddings(
                chunk.content, self.embedding_model
            )
            self._chunks.append(chunk)
            self._embeddings.append(embedding)

    async def async_insert(self, chunks: List[Chunk]) -> None:
        for chunk in chunks:
            embedding = await self.embedding_interface.async_get_embeddings(
                chunk.content, self.embedding_model
            )
            self._chunks.append(chunk)
            self._embeddings.append(embedding)

    def sync_get_num_documents(self) -> int:
        return len(self._chunks)

    async def async_get_num_documents(self) -> int:
        return len(self._chunks)



import os
from typing import Final
from xecai.embeddings.embedding_interface import EmbeddingInterface
from xecai.models import Chunk, SearchType
from xecai.vector_db.vector_db_interface import VectorDBInterface

try:
    import psycopg
    from pgvector.psycopg import register_vector, register_vector_async
    from psycopg_pool import AsyncConnectionPool, ConnectionPool
except ImportError as e:
    raise RuntimeError(
        """PostgreSQL vector DB provider requires the 'postgresql' extra.
        Install with: uv pip install xecai[postgresql]"""
    ) from e


class PostgreSQLVectorDB(VectorDBInterface):
    """
    PostgreSQL vector database implementation using the pgvector extension.

    Since PostgreSQL doesn't inherently generate embeddings from text queries
    like managed services (e.g. AWS Bedrock Knowledge Bases), you must provide
    an embedding interface and model during initialization.
    """

    def __init__(
        self,
        embedding_interface: EmbeddingInterface | None = None,
        embedding_model: str | None = None,
        table_name: str = "vector_documents",
        vector_dim: int = 1536,
    ):
        if not embedding_interface or not embedding_model:
            raise ValueError(
                "embedding_interface and embedding_model must be provided during PostgreSQLVectorDB initialization."
            )

        self.pg_url: Final[str] = os.environ.get(
            "POSTGRESQL_URL", "postgresql://postgres:postgres@localhost:5432/xecai"
        )
        self.table_name: Final[str] = table_name
        self.vector_dim: Final[int] = vector_dim

        self.embedding_interface = embedding_interface
        self.embedding_model = embedding_model

        self._pool: ConnectionPool | None = None
        self._async_pool: AsyncConnectionPool | None = None
        self._ensure_table_exists()

    def _configure_sync_conn(self, conn: psycopg.Connection):
        register_vector(conn)

    async def _configure_async_conn(self, conn: psycopg.AsyncConnection):
        await register_vector_async(conn)

    def _get_pool(self) -> ConnectionPool:
        if self._pool is None:
            self._pool = ConnectionPool(
                self.pg_url, configure=self._configure_sync_conn
            )
        return self._pool

    async def _get_async_pool(self) -> AsyncConnectionPool:
        if self._async_pool is None:
            self._async_pool = AsyncConnectionPool(
                self.pg_url, configure=self._configure_async_conn, open=False
            )
            await self._async_pool.open()
        return self._async_pool

    def _ensure_table_exists(self) -> None:
        query_ext = "CREATE EXTENSION IF NOT EXISTS vector;"
        query_table = f"""
        CREATE TABLE IF NOT EXISTS {self.table_name} (
            id UUID PRIMARY KEY DEFAULT gen_random_uuid(),
            document TEXT NOT NULL,
            origin TEXT,
            fragment INTEGER DEFAULT 0,
            content TEXT NOT NULL,
            embedding vector({self.vector_dim})
        );
        """
        with psycopg.connect(self.pg_url) as conn:
            # Create extension first before registering the vector type
            with conn.cursor() as cur:
                cur.execute(query_ext)
            conn.commit()

            # Now that the type exists in the database, register it with psycopg
            self._configure_sync_conn(conn)

            # Create the table
            with conn.cursor() as cur:
                cur.execute(query_table)
            conn.commit()

    def _build_semantic_query(self) -> str:
        return f"""
            SELECT document, origin, fragment, content, 1 - (embedding <=> %s::vector) AS score
            FROM {self.table_name}
            ORDER BY embedding <=> %s::vector
            LIMIT %s;
        """

    def _build_hybrid_query(self) -> str:
        # Standard Reciprocal Rank Fusion (RRF) for Hybrid Search
        return f"""
            WITH semantic_search AS (
                SELECT id, document, origin, fragment, content,
                       ROW_NUMBER() OVER (ORDER BY embedding <=> %s::vector) AS rank
                FROM {self.table_name}
                ORDER BY embedding <=> %s::vector
                LIMIT 100
            ),
            keyword_search AS (
                SELECT id, document, origin, fragment, content,
                       ROW_NUMBER() OVER (ORDER BY ts_rank_cd(to_tsvector('english', content), plainto_tsquery('english', %s)) DESC) AS rank
                FROM {self.table_name}
                WHERE to_tsvector('english', content) @@ plainto_tsquery('english', %s)
                ORDER BY rank
                LIMIT 100
            )
            SELECT COALESCE(s.document, k.document) as document,
                   COALESCE(s.origin, k.origin) as origin,
                   COALESCE(s.fragment, k.fragment) as fragment,
                   COALESCE(s.content, k.content) as content,
                   COALESCE(1.0 / (60 + s.rank), 0.0) + COALESCE(1.0 / (60 + k.rank), 0.0) AS rrf_score
            FROM semantic_search s
            FULL OUTER JOIN keyword_search k ON s.id = k.id
            ORDER BY rrf_score DESC
            LIMIT %s;
        """

    def _map_row_to_chunk(self, row: tuple) -> Chunk:
        return Chunk(
            document=row[0],
            origin=row[1],
            fragment=row[2] or 0,
            content=row[3],
            score=row[4] if len(row) > 4 else None,
        )

    def sync_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> list[Chunk]:
        query_embedding = self.embedding_interface.sync_get_embeddings(
            query, self.embedding_model
        )

        results = []
        with self._get_pool().connection() as conn:
            with conn.cursor() as cur:
                if search_type == SearchType.SEMANTIC:
                    cur.execute(
                        self._build_semantic_query(),
                        (query_embedding, query_embedding, k),
                    )
                    for row in cur.fetchall():
                        results.append(self._map_row_to_chunk(row))
                elif search_type == SearchType.HYBRID:
                    cur.execute(
                        self._build_hybrid_query(),
                        (query_embedding, query_embedding, query, query, k),
                    )
                    for row in cur.fetchall():
                        results.append(self._map_row_to_chunk(row))

        return results

    async def async_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> list[Chunk]:
        query_embedding = await self.embedding_interface.async_get_embeddings(
            query, self.embedding_model
        )

        results = []
        pool = await self._get_async_pool()
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                if search_type == SearchType.SEMANTIC:
                    await cur.execute(
                        self._build_semantic_query(),
                        (query_embedding, query_embedding, k),
                    )
                    for row in await cur.fetchall():
                        results.append(self._map_row_to_chunk(row))
                elif search_type == SearchType.HYBRID:
                    await cur.execute(
                        self._build_hybrid_query(),
                        (query_embedding, query_embedding, query, query, k),
                    )
                    for row in await cur.fetchall():
                        results.append(self._map_row_to_chunk(row))

        return results

    def sync_insert(self, chunks: list[Chunk]) -> None:
        with self._get_pool().connection() as conn:
            with conn.cursor() as cur:
                for chunk in chunks:
                    embedding = self.embedding_interface.sync_get_embeddings(
                        chunk.content, self.embedding_model
                    )
                    cur.execute(
                        f"""
                        INSERT INTO {self.table_name} (document, origin, fragment, content, embedding)
                        VALUES (%s, %s, %s, %s, %s::vector)
                        """,
                        (
                            chunk.document,
                            chunk.origin,
                            chunk.fragment,
                            chunk.content,
                            embedding,
                        ),
                    )
            conn.commit()

    async def async_insert(self, chunks: list[Chunk]) -> None:
        pool = await self._get_async_pool()
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                for chunk in chunks:
                    embedding = await self.embedding_interface.async_get_embeddings(
                        chunk.content, self.embedding_model
                    )
                    await cur.execute(
                        f"""
                        INSERT INTO {self.table_name} (document, origin, fragment, content, embedding)
                        VALUES (%s, %s, %s, %s, %s::vector)
                        """,
                        (
                            chunk.document,
                            chunk.origin,
                            chunk.fragment,
                            chunk.content,
                            embedding,
                        ),
                    )
            await conn.commit()

    def sync_get_num_documents(self) -> int:
        with self._get_pool().connection() as conn:
            with conn.cursor() as cur:
                cur.execute(f"SELECT COUNT(*) FROM {self.table_name}")
                return cur.fetchone()[0]

    async def async_get_num_documents(self) -> int:
        pool = await self._get_async_pool()
        async with pool.connection() as conn:
            async with conn.cursor() as cur:
                await cur.execute(f"SELECT COUNT(*) FROM {self.table_name}")
                return (await cur.fetchone())[0]



import os
from typing import Final
from xecai.memory.memory_interface import MemoryInterface
from xecai.models import Conversation

try:
    import redis
    import redis.asyncio as redis_async
except ImportError as e:
    raise RuntimeError(
        """Redis memory provider requires the 'redis' extra.
        Install with: uv pip install xecai[redis]"""
    ) from e


class RedisMemory(MemoryInterface):
    def __init__(self):
        self.redis_url: Final[str] = os.environ.get(
            "REDIS_URL", "redis://127.0.0.1:6379/0"
        )
        self.client: Final[redis.Redis] = redis.from_url(self.redis_url)
        self.async_client: Final[redis_async.Redis] = redis_async.from_url(
            self.redis_url
        )
        self.prefix: Final[str] = "conversation:"

    def _get_key(self, conversation_id: str) -> str:
        return f"{self.prefix}{conversation_id}"

    def sync_get_conversation(self, conversation_id: str) -> Conversation | None:
        data = self.client.get(self._get_key(conversation_id))
        if data:
            return Conversation.model_validate_json(data)
        return None

    async def async_get_conversation(self, conversation_id: str) -> Conversation | None:
        data = await self.async_client.get(self._get_key(conversation_id))
        if data:
            return Conversation.model_validate_json(data)
        return None

    def sync_save_conversation(self, conversation: Conversation) -> None:
        key = self._get_key(str(conversation.conversation_id))
        self.client.set(key, conversation.model_dump_json())

    async def async_save_conversation(self, conversation: Conversation) -> None:
        key = self._get_key(str(conversation.conversation_id))
        await self.async_client.set(key, conversation.model_dump_json())



from datetime import datetime, timezone
from enum import Enum
from uuid import UUID, uuid4
from pydantic import BaseModel, Field


class MessageType(Enum):
    DEVELOPER = "DEVELOPER"
    USER = "USER"
    BOT = "BOT"


class ReasoningOptions(Enum):
    NONE = "none"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class StopReason(Enum):
    END = "end"
    MAX_TOKENS = "max_tokens"
    STOP_SEQUENCE = "stop_sequence"
    TOOL_USE = "tool_use"
    CONTENT_FILTER = "content_filter"
    OTHER = "other"


class SearchType(Enum):
    SEMANTIC = "SEMANTIC"
    HYBRID = "HYBRID"


class Message(BaseModel):
    content: str
    message_type: MessageType

    def to_prompt_text(self) -> str:
        return f"{self.message_type.value}: {self.content}"


class Stats(BaseModel):
    input_tokens: int = 0
    output_tokens: int = 0
    total_tokens: int = 0
    cached_input_tokens: int = 0
    cached_output_tokens: int = 0
    reasoning_tokens: int = 0


class ChatResponse(BaseModel):
    text: str
    stats: Stats | None = None
    stop_reason: StopReason | None = None


class Chunk(BaseModel):
    document: str
    origin: str | None = None
    fragment: int = 0
    content: str
    score: float | None = None
    metadata: dict = Field(default_factory=dict)


class Conversation(BaseModel):
    user: str | None = None
    conversation_id: UUID = Field(default_factory=uuid4)
    messages: list[Message] = Field(default_factory=list)
    created_at: datetime = Field(default_factory=lambda: datetime.now(timezone.utc))


class CredentialsError(Exception):
    pass


class RateLimitError(Exception):
    pass


class BadRequestError(Exception):
    pass


class ModelNotFoundError(Exception):
    pass



from abc import ABC, abstractmethod
from xecai.models import Conversation


class MemoryInterface(ABC):
    @abstractmethod
    def sync_get_conversation(self, conversation_id: str) -> Conversation | None:
        pass

    @abstractmethod
    async def async_get_conversation(self, conversation_id: str) -> Conversation | None:
        pass

    @abstractmethod
    def sync_save_conversation(self, conversation: Conversation) -> None:
        pass

    @abstractmethod
    async def async_save_conversation(self, conversation: Conversation) -> None:
        pass



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Iterator
from typing import Any
from xecai.models import ChatResponse, Message, ReasoningOptions


class ChatInterface(ABC):
    @abstractmethod
    def check_model(self, model_name: str) -> None:
        pass

    @staticmethod
    @abstractmethod
    def messages_to_custom_messages(system_prompt: str, messages: list[Message]) -> Any:
        pass

    @abstractmethod
    def invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        pass

    @abstractmethod
    async def async_invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        pass

    @abstractmethod
    def stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> Iterator[ChatResponse]:
        yield ChatResponse(text="")

    @abstractmethod
    async def async_stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> AsyncIterator[ChatResponse]:
        yield ChatResponse(text="")



import functools
from collections.abc import AsyncIterator, Awaitable, Iterator
from typing import Callable, NoReturn, ParamSpec, TypeVar
from xecai.models import BadRequestError, RateLimitError


try:
    from google.genai.errors import APIError
except ImportError as e:
    raise RuntimeError(
        """Google provider requires the 'google' extra.
        Install with: uv pip install xecai[google]"""
    ) from e

P = ParamSpec("P")
T = TypeVar("T")


def handle_api_error(e: APIError) -> NoReturn:
    code = None
    if isinstance(e.code, int):
        code = e.code
    elif getattr(e, "details", None) and isinstance(e.details, dict):
        error_details = e.details.get("error", {})
        if isinstance(error_details, dict):
            code = error_details.get("code")

    message = getattr(e, "message", str(e))
    if code == 400:
        raise BadRequestError(message) from e
    elif code == 429 or code == 500 or code == 503:
        raise RateLimitError(message) from e

    raise e


def sync_error_decorator(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return func(*args, **kwargs)
        except APIError as e:
            handle_api_error(e)

    return wrapper


def async_error_decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return await func(*args, **kwargs)
        except APIError as e:
            handle_api_error(e)

    return wrapper


def sync_generator_error_decorator(
    func: Callable[P, Iterator[T]],
) -> Callable[P, Iterator[T]]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Iterator[T]:
        try:
            yield from func(*args, **kwargs)
        except APIError as e:
            handle_api_error(e)

    return wrapper


def async_generator_error_decorator(
    func: Callable[P, AsyncIterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> AsyncIterator[T]:
        try:
            async for item in func(*args, **kwargs):
                yield item
        except APIError as e:
            handle_api_error(e)

    return wrapper



import functools
from collections.abc import AsyncIterator, Awaitable, Iterator
from typing import Callable, ParamSpec, TypeVar
from xecai.models import BadRequestError, RateLimitError

try:
    import anthropic
except ImportError as e:
    raise RuntimeError(
        """Anthropic provider requires the 'anthropic' extra.
        Install with: uv pip install xecai[anthropic]"""
    ) from e


P = ParamSpec("P")
T = TypeVar("T")


def sync_error_decorator(func: Callable[P, T]) -> Callable[P, T]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return func(*args, **kwargs)
        except anthropic.BadRequestError as e:
            raise BadRequestError(str(e)) from e
        except anthropic.RateLimitError as e:
            raise RateLimitError(str(e)) from e
        except anthropic.InternalServerError as e:
            raise RateLimitError(str(e)) from e

    return wrapper


def async_error_decorator(func: Callable[P, Awaitable[T]]) -> Callable[P, Awaitable[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            return await func(*args, **kwargs)
        except anthropic.BadRequestError as e:
            raise BadRequestError(str(e)) from e
        except anthropic.RateLimitError as e:
            raise RateLimitError(str(e)) from e
        except anthropic.InternalServerError as e:
            raise RateLimitError(str(e)) from e

    return wrapper


def sync_generator_error_decorator(
    func: Callable[P, Iterator[T]],
) -> Callable[P, Iterator[T]]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Iterator[T]:
        try:
            yield from func(*args, **kwargs)
        except anthropic.BadRequestError as e:
            raise BadRequestError(str(e)) from e
        except anthropic.RateLimitError as e:
            raise RateLimitError(str(e)) from e
        except anthropic.InternalServerError as e:
            raise RateLimitError(str(e)) from e

    return wrapper


def async_generator_error_decorator(
    func: Callable[P, AsyncIterator[T]],
) -> Callable[P, AsyncIterator[T]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> AsyncIterator[T]:
        try:
            async for item in func(*args, **kwargs):
                yield item
        except anthropic.BadRequestError as e:
            raise BadRequestError(str(e)) from e
        except anthropic.RateLimitError as e:
            raise RateLimitError(str(e)) from e
        except anthropic.InternalServerError as e:
            raise RateLimitError(str(e)) from e

    return wrapper



import os
from typing import Final, List, Dict, Any

from xecai.models import Chunk
from xecai.reranker.reranker_interface import RerankerInterface
from xecai.error_handlers.aws_error_handler import (
    sync_error_decorator,
    async_error_decorator,
)

try:
    import boto3
    from aiobotocore.session import AioSession, get_session
except ImportError as e:
    raise RuntimeError(
        "AWS Bedrock provider requires the 'aws' extra.\n"
        "Install with: uv pip install xecai[aws]"
    ) from e


class AWSReranker(RerankerInterface):
    _SERVICE_NAME: Final[str] = "bedrock-agent-runtime"

    def __init__(
        self,
        region_name: str | None = None,
        model_arn: str | None = None,
    ) -> None:
        self.region_name: Final[str] = region_name or os.getenv("AWS_REGION", "us-east-1")
        default_model = (
            f"arn:aws:bedrock:{self.region_name}"
            "::foundation-model/cohere.rerank-v3-5:0"
        )
        self.model_arn: Final[str] = model_arn or os.getenv(
            "AWS_RERANK_MODEL_ARN",
            default_model,
        )
        self.client: Final = boto3.client(
            self._SERVICE_NAME,
            region_name=self.region_name,
        )
        self._aiosession: Final[AioSession] = get_session()

    @staticmethod
    def _prepare_sources(chunks: List[Chunk]) -> List[Dict[str, Any]]:
        return [
            {
                "type": "INLINE",
                "inlineDocumentSource": {
                    "type": "TEXT",
                    "textDocument": {"text": chunk.content},
                },
            }
            for chunk in chunks
        ]

    @staticmethod
    def _map_response(response: Dict[str, Any], chunks: List[Chunk]) -> List[Chunk]:
        results = response.get("results", [])
        reranked: List[Chunk] = []

        for result in results:
            idx = result.get("index")
            score = result.get("relevanceScore")

            if idx is not None and 0 <= idx < len(chunks):
                reranked.append(
                    chunks[idx].model_copy(update={"score": score})
                )

        return reranked

    def _build_request(
        self,
        query: str,
        chunks: List[Chunk],
        k: int,
    ) -> Dict[str, Any]:

        return {
            "queries": [
                {
                    "type": "TEXT",
                    "textQuery": {"text": query},
                }
            ],
            "sources": self._prepare_sources(chunks),
            "rerankingConfiguration": {
                "type": "BEDROCK_RERANKING_MODEL",
                "bedrockRerankingConfiguration": {
                    "numberOfResults": min(k, len(chunks)),
                    "modelConfiguration": {"modelArn": self.model_arn},
                },
            },
        }


    @sync_error_decorator
    def sync_rerank(
        self,
        query: str,
        chunks: List[Chunk],
        k: int,
    ) -> List[Chunk]:
        if not chunks:
            return []

        request = self._build_request(query, chunks, k)
        response = self.client.rerank(**request)
        return self._map_response(response, chunks)


    @async_error_decorator
    async def async_rerank(
        self,
        query: str,
        chunks: List[Chunk],
        k: int,
    ) -> List[Chunk]:

        if not chunks:
            return []

        request = self._build_request(query, chunks, k)
        async with self._aiosession.create_client(
            self._SERVICE_NAME,
            region_name=self.region_name,
        ) as client:
            response = await client.rerank(**request)

        return self._map_response(response, chunks)



import os
from collections.abc import AsyncIterator, Iterator
from typing import Final
from xecai.chat.chat_interface import ChatInterface
from xecai.models import (
    ChatResponse,
    Message,
    MessageType,
    ReasoningOptions,
    Stats,
    StopReason,
    ModelNotFoundError,
)

from xecai.error_handlers.aws_error_handler import (
    async_error_decorator,
    async_generator_error_decorator,
    sync_error_decorator,
    sync_generator_error_decorator,
)

try:
    import boto3
    from aiobotocore.session import AioSession, get_session
except ImportError as e:
    raise RuntimeError(
        """AWS Bedrock provider requires the 'aws' extra.
        Install with: uv pip install xecai[aws]"""
    ) from e


class AWSChat(ChatInterface):
    def __init__(self):
        self.region_name: Final[str] = os.environ.get("AWS_REGION", "us-east-1")
        self.client: Final[boto3.session.Session.client] = boto3.client(
            "bedrock-runtime", region_name=self.region_name
        )
        self.bedrock_client: Final[boto3.session.Session.client] = boto3.client(
            "bedrock", region_name=self.region_name
        )
        self.aiosession: Final[AioSession] = get_session()

    @sync_error_decorator
    def check_model(self, model_name: str) -> None:
        response = self.bedrock_client.list_foundation_models()
        available_models = response.get("modelSummaries", [])
        if any(m.get("modelId") == model_name for m in available_models):
            return

        try:
            profiles_response = self.bedrock_client.list_inference_profiles()
            available_profiles = profiles_response.get("inferenceProfileSummaries", [])
            if any(p.get("inferenceProfileId") == model_name for p in available_profiles):
                return
        except Exception:
            pass

        raise ModelNotFoundError(f"Model '{model_name}' not found.")

    @staticmethod
    def messages_to_custom_messages(
        system_prompt: str, messages: list[Message]
    ) -> list:
        contents = []
        for msg in messages:
            if msg.message_type == MessageType.BOT:
                role = "assistant"
            else:
                role = "user"

            contents.append({"role": role, "content": [{"text": msg.content}]})
        return contents

    def _prepare_kwargs(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None,
        temperature: float | None,
        retries: int = 3,
    ) -> dict:

        kwargs = {
            "modelId": model_name,
            "messages": self.messages_to_custom_messages(system_prompt, messages),
            "system": [{"text": system_prompt}],
        }

        if temperature is not None:
            kwargs["inferenceConfig"] = {"temperature": temperature}

        if reasoning and reasoning != ReasoningOptions.NONE:
            budget_map = {
                ReasoningOptions.LOW: 1024,
                ReasoningOptions.MEDIUM: 4096,
                ReasoningOptions.HIGH: 16384,
            }
            budget = budget_map[reasoning]
            kwargs["additionalModelRequestFields"] = {
                "reasoning_config": {"type": "enabled", "budget_tokens": budget}
            }

        return kwargs

    def _map_stop_reason(self, stop_reason: str | None) -> StopReason:
        if stop_reason == "end_turn":
            return StopReason.END
        elif stop_reason == "max_tokens":
            return StopReason.MAX_TOKENS
        elif stop_reason == "stop_sequence":
            return StopReason.STOP_SEQUENCE
        elif stop_reason == "tool_use":
            return StopReason.TOOL_USE
        elif stop_reason == "content_filtered":
            return StopReason.CONTENT_FILTER
        elif stop_reason is None:
            return StopReason.END
        return StopReason.OTHER

    @sync_error_decorator
    def invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        response = self.client.converse(**kwargs)

        usage = response.get("usage", {})
        stats = Stats(
            input_tokens=usage.get("inputTokens", 0) or 0,
            output_tokens=usage.get("outputTokens", 0) or 0,
            total_tokens=usage.get("totalTokens", 0) or 0,
        )

        text = (
            response.get("output", {})
            .get("message", {})
            .get("content", [{}])[0]
            .get("text", "")
        )

        stop_reason = response.get("stopReason")

        return ChatResponse(
            text=text, stats=stats, stop_reason=self._map_stop_reason(stop_reason)
        )

    @async_error_decorator
    async def async_invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        async with self.aiosession.create_client(
            "bedrock-runtime", region_name=self.region_name
        ) as client:
            response = await client.converse(**kwargs)

            usage = response.get("usage", {})
            stats = Stats(
                input_tokens=usage.get("inputTokens", 0) or 0,
                output_tokens=usage.get("outputTokens", 0) or 0,
                total_tokens=usage.get("totalTokens", 0) or 0,
            )

            text = (
                response.get("output", {})
                .get("message", {})
                .get("content", [{}])[0]
                .get("text", "")
            )

            stop_reason = response.get("stopReason")

            return ChatResponse(
                text=text, stats=stats, stop_reason=self._map_stop_reason(stop_reason)
            )

    @sync_generator_error_decorator
    def stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> Iterator[ChatResponse]:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        response = self.client.converse_stream(**kwargs)
        for chunk in response.get("stream", []):
            if "contentBlockDelta" in chunk:
                delta = chunk["contentBlockDelta"].get("delta", {})
                if "text" in delta:
                    yield ChatResponse(text=delta["text"])
            elif "messageStop" in chunk:
                stop_reason = chunk["messageStop"].get("stopReason")
                yield ChatResponse(
                    text="", stop_reason=self._map_stop_reason(stop_reason)
                )
            elif "metadata" in chunk:
                usage = chunk["metadata"].get("usage", {})
                if usage:
                    stats = Stats(
                        input_tokens=usage.get("inputTokens", 0) or 0,
                        output_tokens=usage.get("outputTokens", 0) or 0,
                        total_tokens=usage.get("totalTokens", 0) or 0,
                    )
                    yield ChatResponse(text="", stats=stats)

    @async_generator_error_decorator
    async def async_stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> AsyncIterator[ChatResponse]:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        async with self.aiosession.create_client(
            "bedrock-runtime", region_name=self.region_name
        ) as client:
            response = await client.converse_stream(**kwargs)
            stream = response.get("stream")
            if stream:
                async for chunk in stream:
                    if "contentBlockDelta" in chunk:
                        delta = chunk["contentBlockDelta"].get("delta", {})
                        if "text" in delta:
                            yield ChatResponse(text=delta["text"])
                    elif "messageStop" in chunk:
                        stop_reason = chunk["messageStop"].get("stopReason")
                        yield ChatResponse(
                            text="", stop_reason=self._map_stop_reason(stop_reason)
                        )
                    elif "metadata" in chunk:
                        usage = chunk["metadata"].get("usage", {})
                        if usage:
                            stats = Stats(
                                input_tokens=usage.get("inputTokens", 0) or 0,
                                output_tokens=usage.get("outputTokens", 0) or 0,
                                total_tokens=usage.get("totalTokens", 0) or 0,
                            )
                            yield ChatResponse(text="", stats=stats)



import os
from collections.abc import AsyncIterator, Iterator
from typing import Any, Final
from xecai.chat.chat_interface import ChatInterface
from xecai.models import (
    ChatResponse,
    Message,
    MessageType,
    ReasoningOptions,
    Stats,
    StopReason,
    ModelNotFoundError,
)
from xecai.error_handlers.openai_error_handler import (
    async_error_decorator,
    async_generator_error_decorator,
    sync_error_decorator,
    sync_generator_error_decorator,
)

try:
    import openai
except ImportError as e:
    raise RuntimeError(
        """OpenAI provider requires the 'openai' extra.
        Install with: uv pip install xecai[openai]"""
    ) from e


class OpenAIChat(ChatInterface):
    def __init__(self):
        api_key = os.environ.get("OPENAI_API_KEY")
        self.client: Final[openai.OpenAI] = openai.OpenAI(api_key=api_key)
        self.aclient: Final[openai.AsyncOpenAI] = openai.AsyncOpenAI(api_key=api_key)

    @staticmethod
    def messages_to_custom_messages(
        system_prompt: str, messages: list[Message]
    ) -> list[dict]:
        inputs = []

        if system_prompt:
            inputs.append(
                {
                    "role": "developer",
                    "content": [{"type": "input_text", "text": system_prompt}],
                }
            )

        for msg in messages:
            if msg.message_type == MessageType.BOT:
                role = "assistant"
            elif msg.message_type == MessageType.DEVELOPER:
                role = "developer"
            else:
                role = "user"

            inputs.append(
                {
                    "role": role,
                    "content": [{"type": "input_text", "text": msg.content}],
                }
            )

        return inputs

    @sync_error_decorator
    def check_model(self, model_name: str) -> None:
        available_models = self.client.models.list()
        if not any(m.id == model_name for m in available_models.data):
            raise ModelNotFoundError(f"Model '{model_name}' not found.")

    def _prepare_kwargs(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None,
        temperature: float | None,
        stream: bool = False,
        retries: int = 3,
    ) -> dict:
        kwargs = {
            "model": model_name,
            "input": self.messages_to_custom_messages(system_prompt, messages),
        }

        if stream:
            kwargs["stream"] = True

        if temperature is not None:
            kwargs["temperature"] = temperature

        if reasoning and reasoning != ReasoningOptions.NONE:
            kwargs["reasoning"] = {"effort": reasoning.value}

        return kwargs

    def _map_stop_reason(
        self, status: str | None, incomplete_details: Any | None
    ) -> StopReason:
        if status == "completed":
            return StopReason.END
        elif status == "incomplete":
            reason = (
                getattr(incomplete_details, "reason", None)
                if incomplete_details
                else None
            )
            if reason == "max_output_tokens":
                return StopReason.MAX_TOKENS
            elif reason == "content_filter":
                return StopReason.CONTENT_FILTER
            elif reason == "stop":
                return StopReason.STOP_SEQUENCE
        return StopReason.OTHER

    @sync_error_decorator
    def invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        response = self.client.responses.create(**kwargs)
        usage = response.usage
        stats = (
            Stats(
                input_tokens=getattr(usage, "input_tokens", 0) or 0,
                output_tokens=getattr(usage, "output_tokens", 0) or 0,
                total_tokens=getattr(usage, "total_tokens", 0) or 0,
                cached_input_tokens=getattr(
                    getattr(usage, "input_tokens_details", None), "cached_tokens", 0
                )
                or 0,
                reasoning_tokens=getattr(
                    getattr(usage, "output_tokens_details", None), "reasoning_tokens", 0
                )
                or 0,
            )
            if usage
            else Stats()
        )

        status = response.status
        if status == "completed" and response.output:
            status = response.output[0].status

        stop_reason = self._map_stop_reason(
            status, getattr(response, "incomplete_details", None)
        )

        return ChatResponse(
            text=response.output_text or "", stats=stats, stop_reason=stop_reason
        )

    @async_error_decorator
    async def async_invoke(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> ChatResponse:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
        )
        response = await self.aclient.responses.create(**kwargs)

        usage = response.usage
        stats = (
            Stats(
                input_tokens=getattr(usage, "input_tokens", 0) or 0,
                output_tokens=getattr(usage, "output_tokens", 0) or 0,
                total_tokens=getattr(usage, "total_tokens", 0) or 0,
                cached_input_tokens=getattr(
                    getattr(usage, "input_tokens_details", None), "cached_tokens", 0
                )
                or 0,
                reasoning_tokens=getattr(
                    getattr(usage, "output_tokens_details", None), "reasoning_tokens", 0
                )
                or 0,
            )
            if usage
            else Stats()
        )

        status = response.status
        if status == "completed" and response.output:
            status = response.output[0].status

        stop_reason = self._map_stop_reason(
            status, getattr(response, "incomplete_details", None)
        )

        return ChatResponse(
            text=response.output_text or "", stats=stats, stop_reason=stop_reason
        )

    @sync_generator_error_decorator
    def stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> Iterator[ChatResponse]:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
            stream=True,
        )
        response = self.client.responses.create(**kwargs)

        for event in response:
            if event.type == "response.output_text.delta":
                yield ChatResponse(text=event.delta)
            elif event.type == "response.completed" and hasattr(event, "response"):
                usage = getattr(event.response, "usage", None)
                stats = (
                    Stats(
                        input_tokens=getattr(usage, "input_tokens", 0) or 0,
                        output_tokens=getattr(usage, "output_tokens", 0) or 0,
                        total_tokens=getattr(usage, "total_tokens", 0) or 0,
                        cached_input_tokens=getattr(
                            getattr(usage, "input_tokens_details", None),
                            "cached_tokens",
                            0,
                        )
                        or 0,
                        reasoning_tokens=getattr(
                            getattr(usage, "output_tokens_details", None),
                            "reasoning_tokens",
                            0,
                        )
                        or 0,
                    )
                    if usage
                    else Stats()
                )

                status = getattr(event.response, "status", None)
                stop_reason = self._map_stop_reason(
                    status, getattr(event.response, "incomplete_details", None)
                )

                yield ChatResponse(text="", stats=stats, stop_reason=stop_reason)

    @async_generator_error_decorator
    async def async_stream(
        self,
        model_name: str,
        system_prompt: str,
        messages: list[Message],
        reasoning: ReasoningOptions | None = None,
        temperature: float | None = None,
        retries: int = 3,
    ) -> AsyncIterator[ChatResponse]:
        kwargs = self._prepare_kwargs(
            model_name,
            system_prompt,
            messages,
            reasoning,
            temperature,
            stream=True,
        )
        response = await self.aclient.responses.create(**kwargs)

        async for event in response:
            if event.type == "response.output_text.delta":
                yield ChatResponse(text=event.delta)
            elif event.type == "response.completed" and hasattr(event, "response"):
                usage = getattr(event.response, "usage", None)
                stats = (
                    Stats(
                        input_tokens=getattr(usage, "input_tokens", 0) or 0,
                        output_tokens=getattr(usage, "output_tokens", 0) or 0,
                        total_tokens=getattr(usage, "total_tokens", 0) or 0,
                        cached_input_tokens=getattr(
                            getattr(usage, "input_tokens_details", None),
                            "cached_tokens",
                            0,
                        )
                        or 0,
                        reasoning_tokens=getattr(
                            getattr(usage, "output_tokens_details", None),
                            "reasoning_tokens",
                            0,
                        )
                        or 0,
                    )
                    if usage
                    else Stats()
                )

                status = getattr(event.response, "status", None)
                stop_reason = self._map_stop_reason(
                    status, getattr(event.response, "incomplete_details", None)
                )

                yield ChatResponse(text="", stats=stats, stop_reason=stop_reason)



import json
import os
from typing import Any, Final
from xecai.embeddings.embedding_interface import EmbeddingInterface
from xecai.error_handlers.aws_error_handler import async_error_decorator, sync_error_decorator

try:
    import boto3
    from aiobotocore.session import AioSession, get_session
except ImportError as e:
    raise RuntimeError(
        """AWS Bedrock provider requires the 'aws' extra.
        Install with: uv pip install xecai[aws]"""
    ) from e


class AWSEmbedding(EmbeddingInterface):
    def __init__(self):
        self.region_name: Final[str] = os.environ.get("AWS_REGION", "us-east-1")
        self.client: Final[boto3.session.Session.client] = boto3.client(
            "bedrock-runtime", region_name=self.region_name
        )
        self.aiosession: Final[AioSession] = get_session()

    def _invoke_sync(self, model: str, body: dict) -> dict[str, Any]:
        response = self.client.invoke_model(
            modelId=model,
            body=json.dumps(body),
            accept="application/json",
            contentType="application/json",
        )
        return json.loads(response.get("body").read())

    async def _invoke_async(
        self, client: Any, model: str, body: dict
    ) -> dict[str, Any]:
        response = await client.invoke_model(
            modelId=model,
            body=json.dumps(body),
            accept="application/json",
            contentType="application/json",
        )
        return json.loads(await response.get("body").read())

    @sync_error_decorator
    def sync_get_embeddings(self, text: str, model: str) -> list[float]:
        if "cohere" in model.lower():
            body = {"texts": [text], "input_type": "search_document"}
            response_body = self._invoke_sync(model, body)
            embeddings = response_body.get("embeddings", [])
            return embeddings[0] if embeddings else []

        elif "titan" in model.lower():
            body = {"inputText": text}
            response_body = self._invoke_sync(model, body)
            return response_body.get("embedding", [])

        else:
            raise ValueError(f"Unsupported AWS embedding model family: {model}")

    @async_error_decorator
    async def async_get_embeddings(self, text: str, model: str) -> list[float]:
        async with self.aiosession.create_client(
            "bedrock-runtime", region_name=self.region_name
        ) as client:
            if "cohere" in model.lower():
                body = {"texts": [text], "input_type": "search_document"}
                response_body = await self._invoke_async(client, model, body)
                embeddings = response_body.get("embeddings", [])
                return embeddings[0] if embeddings else []

            elif "titan" in model.lower():
                body = {"inputText": text}
                response_body = await self._invoke_async(client, model, body)
                return response_body.get("embedding", [])

            else:
                raise ValueError(f"Unsupported AWS embedding model family: {model}")



from xecai.chat.chat_interface import ChatInterface
from xecai.memory.memory_interface import MemoryInterface
from xecai.models import Conversation, Message, MessageType

DEFAULT_SUMMARY_PROMPT = (
    "Summarize the following conversation. Focus on the main topics and context."
)


def _prepare_summary_payload(conversation: Conversation) -> list[Message]:
    """Prepares the message payload for the summary LLM. Returns None if no conversation exists."""
    conversation_text = "\n".join(
        [msg.to_prompt_text() for msg in conversation.messages]
    )
    return [Message(message_type=MessageType.USER, content=conversation_text)]


def sync_summarize_conversation(
    memory: MemoryInterface,
    chat: ChatInterface,
    conversation_id: str,
    summary_model_name: str,
    summary_prompt: str = DEFAULT_SUMMARY_PROMPT,
) -> None:
    conversation = memory.sync_get_conversation(conversation_id)
    if not conversation or not conversation.messages:
        return

    messages = _prepare_summary_payload(conversation)
    response = chat.invoke(
        model_name=summary_model_name,
        system_prompt=summary_prompt,
        messages=messages,
    )

    conversation.messages = [
        Message(message_type=MessageType.BOT, content=response.text)
    ]
    memory.sync_save_conversation(conversation)


async def async_summarize_conversation(
    memory: MemoryInterface,
    chat: ChatInterface,
    conversation_id: str,
    summary_model_name: str,
    summary_prompt: str = DEFAULT_SUMMARY_PROMPT,
) -> None:
    conversation = await memory.async_get_conversation(conversation_id)
    if not conversation or not conversation.messages:
        return

    messages = _prepare_summary_payload(conversation)
    response = await chat.async_invoke(
        model_name=summary_model_name,
        system_prompt=summary_prompt,
        messages=messages,
    )

    conversation.messages = [
        Message(message_type=MessageType.BOT, content=response.text)
    ]
    await memory.async_save_conversation(conversation)



import logging
from typing import Dict
from xecai.memory.memory_interface import MemoryInterface
from xecai.models import Conversation


class DebugMemory(MemoryInterface):
    """
    Debug implementation of MemoryInterface that stores conversations in an in-memory dictionary.
    """

    def __init__(self):
        logging.warning(
            "DebugMemory is for debugging purposes only and is not recommended for production usecases."
        )
        self._storage: Dict[str, Conversation] = {}

    def sync_get_conversation(self, conversation_id: str) -> Conversation | None:
        return self._storage.get(str(conversation_id))

    async def async_get_conversation(self, conversation_id: str) -> Conversation | None:
        return self.sync_get_conversation(conversation_id)

    def sync_save_conversation(self, conversation: Conversation) -> None:
        self._storage[str(conversation.conversation_id)] = conversation

    async def async_save_conversation(self, conversation: Conversation) -> None:
        self.sync_save_conversation(conversation)



from abc import ABC, abstractmethod
from xecai.models import Chunk, SearchType


class VectorDBInterface(ABC):
    @abstractmethod
    def sync_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> list[Chunk]:
        pass

    @abstractmethod
    async def async_retrieve(
        self, query: str, k: int, search_type: SearchType = SearchType.SEMANTIC
    ) -> list[Chunk]:
        pass

    @abstractmethod
    def sync_insert(self, chunks: list[Chunk]) -> None:
        pass

    @abstractmethod
    async def async_insert(self, chunks: list[Chunk]) -> None:
        pass

    @abstractmethod
    def sync_get_num_documents(self) -> int:
        pass

    @abstractmethod
    async def async_get_num_documents(self) -> int:
        pass



import pytest
from xecai.models import Message, MessageType

def test_message_creation():
    msg = Message(content="hello", message_type=MessageType.USER)
    assert msg.content == "hello"
    assert msg.message_type == MessageType.USER
